#!/usr/bin/python

import wx, os, os.path, glob, re
from wx import xrc

pw, ph = 500.0, 500.0 # picture frame size
tw, th = 1300.0, 1030.0 # tiff file size
scale = min(pw/tw, ph/th)

class CensorApp(wx.App):
    def OnInit(self):
        # sets up the application, loads the layout, initializes it
        # and hooks it together

        # what state variables do we need?
        self.state = {}
        self.state['dirty'] = False # have we censored anything since our last save?
        self.state['cpdata'] = [] # list of tuples of form (filename, [(x1 y1), (x2 y2)]).
                                  # x1 y1 are IMAGE COORDINATES. we never write to cpdata.
        self.state['index'] = 0 # where we are in the cpdata list
        self.state['remnants'] = [] # identical to cpdata, except we subtract from here
                                    # here, x1 y1 are WINDOW COORDINATES.
        self.state['rects'] = [ [] ] # list of lists of rectangles
        self.state['drag_origin'] = (0,0) # origin of a drag event
        self.state['image'] = None # the 'base' image being currently displayed; always 500x500
        self.state['dragging'] = False # were we just dragging our mouse?

        # load XRC resources
        self.res = xrc.XmlResource('layout.xrc')
        self.frame = self.res.LoadFrame(None, 'MainFrame')

        # bind buttons and menus to their event handlers
        self.frame.Bind(wx.EVT_BUTTON, self.NextPress, id=xrc.XRCID('nextbtn'))
        self.frame.Bind(wx.EVT_BUTTON, self.PrevPress, id=xrc.XRCID('prevbtn'))
        self.frame.Bind(wx.EVT_BUTTON, self.RevertPress, id=xrc.XRCID('revertbtn'))
        
        self.frame.Bind(wx.EVT_MENU, self.OpenCPDataPress, id=xrc.XRCID('menuopendata'))
        self.frame.Bind(wx.EVT_MENU, self.SavePress, id=xrc.XRCID('menusave'))
        self.frame.Bind(wx.EVT_MENU, self.ExitPress, id=xrc.XRCID('menuexit'))
        self.frame.Bind(wx.EVT_MENU, self.AboutPress, id=xrc.XRCID('menuabout'))

        # grab other controls we'll want to tweak
        self.frame.picture = xrc.XRCCTRL(self.frame, 'picture')
        self.frame.curposCtrl = xrc.XRCCTRL(self.frame, 'curpos')
        self.frame.totalImgsCtrl = xrc.XRCCTRL(self.frame, 'total_imgs')
        self.frame.origObjCtrl = xrc.XRCCTRL(self.frame, 'orig_obj')
        self.frame.remObjCtrl = xrc.XRCCTRL(self.frame, 'rem_obj')

        # don't break while debugging
        # self.state['image'] = wx.ImageFromBitmap(self.frame.picture.GetBitmap())

        # handle click & drag events on the picture
        self.frame.picture.Bind(wx.EVT_MOUSE_EVENTS, self.PictureMouseEvt)

        # size and show the frame
        self.frame.GetSizer().Fit(self.frame)
        self.frame.Show()
        return True

    def NextPress(self, evt):
        if self.state['index'] + 1 < len(self.state['cpdata']):
            self.state['index'] += 1
            self.UpdateLabels()
    
    def PrevPress(self, evt):
        if self.state['index'] > 0:
            self.state['index'] -= 1
            self.UpdateLabels()

    def RevertPress(self, evt):
        i = self.state['index']
        self.state['rects'][i] = []
        fn, objs = self.state['cpdata'][i]
        objs = [(int(x * scale), int(y * scale)) for (x,y) in objs]
        self.state['remnants'][i] = (fn, objs)
        self.UpdateLabels()

    def OpenCPDataPress(self, evt):
        dlg = wx.FileDialog(self.frame, 'Choose a CellProfiler object data TSV file.',
                wildcard = 'Tab-separated values (*.tab, *.xls)|*.tab;*.xls',
                style = wx.FD_OPEN)
        result = dlg.ShowModal()
        if result == wx.ID_CANCEL: return False

        objf = open(dlg.GetPath(), 'r')
        objbuf = objf.readlines()
        objf.close()

        os.chdir(os.path.dirname(dlg.GetPath()))

#       nevermind all this; let's just scrape the image filename from the Set # info.
#        dlg = wx.FileDialog(self.frame, 'Choose a CellProfiler image data TSV file.',
#                wildcard = 'Tab-separated values (*.tab, *.xls)|*.tab,*.xls',
#                style = wx.FD_OPEN)
#        result = dlg.ShowModal()
#        if result == wx.ID_CANCEL: return False

#        imgf = open(dlg.GetPath(), 'r')
#        imgbuf = imgf.readlines()
#        imgf.close()
        r = re.compile("^Set #(?P<set>\d+), (?P<filename>.*)")

        # we're going to assume that Location_Center_X and Location_Center_Y
        # are fields 1 and 2 and not attempt to parse the header.
        objbuf = [line[:-1].split('\t') for line in objbuf[3:]]
        i = -1
        fns = [] # list of filenames for each set
        objs = [] # will be a list of lists of xy's
        for line in objbuf:
            if line[0] != '':
                # we're at the beginning of a new set
                i += 1
                m = r.match(line[0])
                fns.append(m.group('filename'))
                objs.append([])
            objs[i].append( (float(line[1]), float(line[2])) )

        self.state['cpdata'] = zip(fns, objs)

        # go ahead and scale our coordinates into remnants, too
        for q in range(len(fns)):
            objs[q] = [(int(i * scale), int(j * scale)) for (i,j) in objs[q]]
        self.state['remnants'] = zip(fns, objs)

        # and initialize rects
        self.state['rects'] = [ [] for q in fns ]

        self.UpdateLabels()

    def SavePress(self, evt):
        dlg = wx.FileDialog(self.frame, 'Choose a file to write output to.',
                wildcard = 'Tab-separated values (*.tab)|*.tab',
                style = wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        result = dlg.ShowModal()
        if result == wx.ID_CANCEL: return False

        f = open(dlg.GetPath(), 'w')
        print >> f, 'filename\tcounts'
        for q in self.state['remnants']:
            print >> f, q[0], len(q[1])
        f.close()

        dlh = wx.MessageDialog(self.frame, message='Data saved!', style=wx.OK|wx.ICON_INFORMATION)
        dlh.ShowModal()

    def ExitPress(self, evt):
        # maybe worry about being dirty?
        self.frame.Close()

    def AboutPress(self, evt):
        pass

    def UpdateLabels(self):
        i = self.state['index']
        total = len(self.state['cpdata'][i][1])
        rem = len(self.state['remnants'][i][1])
        
        self.frame.curposCtrl.SetValue("%d" % (i+1))
        self.frame.origObjCtrl.SetValue("%d" % total)
        self.frame.remObjCtrl.SetValue("%d" % rem)
        self.frame.totalImgsCtrl.SetValue("%d" % len(self.state['cpdata']))

        # display the right image, too
        myImage = wx.Image(self.state['cpdata'][i][0])
        w,h = myImage.GetWidth(), myImage.GetHeight()
        myImage = myImage.Scale(int(w*scale), int(h*scale), quality=wx.IMAGE_QUALITY_HIGH)
        imgFrame = wx.EmptyImage(pw,ph,True)
        imgFrame.Paste(myImage,0,0)
        self.state['image'] = imgFrame
        self.frame.picture.SetBitmap(wx.BitmapFromImage(self.state['image']))

        objs = self.state['remnants'][i][1]
        for obj in objs:
            self.state['image'] = wx.ImageFromBitmap(self.DrawCircle(obj))
        self.frame.picture.SetBitmap(wx.BitmapFromImage(self.state['image']))

        for rect in self.state['rects'][i]:
            self.state['image'] = wx.ImageFromBitmap(self.DrawRectangle(rect))

    def DrawRectangle(self, rect):
        tmp = wx.BitmapFromImage(self.state['image'])
        DC = wx.MemoryDC(tmp)
        DC.SetPen(wx.WHITE_PEN)
        DC.SetBrush(wx.TRANSPARENT_BRUSH)
        DC.DrawRectangleRect(rect)
        DC.SelectObject(wx.NullBitmap)
        self.frame.picture.SetBitmap(tmp)
        return tmp

    def DrawCircle(self, obj, display=False):
        tmp = wx.BitmapFromImage(self.state['image'])
        DC = wx.MemoryDC(tmp)
        DC.SetPen(wx.Pen('#0000ff'))
        DC.SetBrush(wx.TRANSPARENT_BRUSH)
        DC.DrawCircle(obj[0], obj[1], 5)
        DC.SelectObject(wx.NullBitmap)
        if display: self.frame.picture.SetBitmap(tmp)
        return tmp

    def PictureMouseEvt(self, evt):
        if evt.ButtonDown():
            self.state['drag_origin'] = evt.GetPositionTuple()
            self.state['dragging'] = False
        if evt.Dragging():
            self.state['dragging'] = True

        tmp, rect = None, None # for scope
        if self.state['dragging']:
            # put down a rectangle
            x,y = self.state['drag_origin']
            xp, yp = evt.GetPositionTuple()
            w,h = xp-x, yp-y
            rect = wx.Rect(x,y,w,h)
            tmp = self.DrawRectangle(rect)
        if evt.ButtonUp():
            if self.state['dragging']:
                # we're done dragging
                self.state['dragging'] = False
                self.state['image'] = wx.ImageFromBitmap(tmp)
                i = self.state['index']
                cpentry = self.state['remnants'][i]
                objs = cpentry[1]
                objs = [obj for obj in objs if not rect.Contains(obj)] # keep all objects outside the rectangle
                # remember that the remnant coordinates are already in window coords
                self.state['remnants'][i] = (cpentry[0], objs)
                self.state['rects'][i].append(rect)
            else: # we just clicked somewhere
                xc, yc = evt.GetPositionTuple()
                self.state['image'] = wx.ImageFromBitmap(self.DrawCircle((xc, yc), True))
                i = self.state['index']
                rems = self.state['remnants'][i]
                objs = rems[1]
                objs.append((xc, yc))
                self.state['remnants'][i] = (rems[0], objs)

            self.frame.remObjCtrl.SetValue("%d" % len(self.state['remnants'][self.state['index']][1]))

def main():
    app = CensorApp(0)
    app.MainLoop()

if __name__ == '__main__':
    main()

